// -*- C++ -*- Allocate exception objects.
// Copyright (C) 2001 Free Software Foundation, Inc.
//
// This file is part of GNU CC.
//
// GNU CC is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// GNU CC is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GNU CC; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

// As a special exception, you may use this file as part of a free software
// library without restriction.  Specifically, if other files instantiate
// templates or use macros or inline functions from this file, or you compile
// this file and link it with other files to produce an executable, this
// file does not by itself cause the resulting executable to be covered by
// the GNU General Public License.  This exception does not however
// invalidate any other reasons why the executable file might be covered by
// the GNU General Public License.

// This is derived from the C++ ABI for IA-64.  Where we diverge
// for cross-architecture compatibility are noted with "@@@".

#include <exception>
#include "unwind-cxx.h"
#include <l4/util/lock.h>
#include <string.h>

using namespace __cxxabiv1;


// ??? How to control these parameters.

#define EMERGENCY_OBJ_SIZE	1024
#define EMERGENCY_OBJ_COUNT	64

typedef unsigned long bitmask_type;

typedef char one_buffer[EMERGENCY_OBJ_SIZE] __attribute__((aligned));
static one_buffer emergency_buffer[EMERGENCY_OBJ_COUNT];
static bitmask_type emergency_used;
static l4_simple_lock_t emergencylock=0;

/*
 * We always use the emergency-Buffer since we don't want an dynamic
 * memory allocation
 */

extern "C" void *
__cxa_allocate_exception(std::size_t thrown_size)
{
  void *ret;

  thrown_size += sizeof (__cxa_exception);

  l4_simple_lock(&emergencylock);
  bitmask_type used = emergency_used;
  unsigned int which = 0;
  
  if (thrown_size > EMERGENCY_OBJ_SIZE)
    goto failed;
  while (used & 1)
  {
    used >>= 1;
    if (++which >= EMERGENCY_OBJ_COUNT)
      goto failed;
  }

  emergency_used |= (bitmask_type)1 << which;
  ret = &emergency_buffer[which][0];

 failed:;
  l4_simple_unlock(&emergencylock);
  if (!ret)
  {
    std::terminate ();
  }
  
  memset (ret, 0, sizeof (__cxa_exception));

  return (void *)((char *)ret + sizeof (__cxa_exception));
}


extern "C" void
__cxa_free_exception(void *vptr)
{
  char *ptr = (char *) vptr;
  if (ptr >= &emergency_buffer[0][0]
      && ptr < &emergency_buffer[0][0] + sizeof (emergency_buffer))
    {
      unsigned int which
	= (unsigned)(ptr - &emergency_buffer[0][0]) / EMERGENCY_OBJ_SIZE;

      l4_simple_lock(&emergencylock);
      emergency_used &= ~((bitmask_type)1 << which);
      l4_simple_unlock(&emergencylock);
    }
}
