// -*- C++ -*- Manage the thread-local exception globals.
// Copyright (C) 2001 Free Software Foundation, Inc.
//
// This file is part of GNU CC.
//
// GNU CC is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// GNU CC is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with GNU CC; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

// As a special exception, you may use this file as part of a free software
// library without restriction.  Specifically, if other files instantiate
// templates or use macros or inline functions from this file, or you compile
// this file and link it with other files to produce an executable, this
// file does not by itself cause the resulting executable to be covered by
// the GNU General Public License.  This exception does not however
// invalidate any other reasons why the executable file might be covered by
// the GNU General Public License.


#include <exception>
#include "l4cpp-config.h"
#include "unwind-cxx.h"
#include <l4/sys/types.h>
#include <l4/sys/syscalls.h>
#include <stdio.h>

using namespace __cxxabiv1;

/*
 * Simple Solution for Thread-Specific storage 
 */
const std::size_t MAX_THREADS=16;
static  __cxa_eh_globals globals_static[MAX_THREADS];

extern "C" __cxa_eh_globals *
__cxa_get_globals_fast ()
{
  l4_threadid_t th = l4_myself();
  size_t i = th.id.lthread;
  if(i>=MAX_THREADS)
  {
    printf("__cxa_get_globals_fast: id=%d>=%d=MAX_THREADS\n",i,MAX_THREADS);
    _exit(1);
  }
  printf("__cxa_eh_globals_fast() i=%d p=%p\n",i,globals_static+i);
  return globals_static+i;
}

extern "C" __cxa_eh_globals *
__cxa_get_globals ()
{
  l4_threadid_t th = l4_myself();
  size_t i = th.id.lthread;
  if(i>=MAX_THREADS)
  {
    printf("__cxa_get_globals: id=%d>=%d=MAX_THREADS\n",i,MAX_THREADS);
    _exit(1);
  }
  printf("__cxa_eh_globals() i=%d p=%p\n",i,globals_static+i);
  return globals_static+i;
}
